import random
import difflib
import time

from tabulate import tabulate
from gaspra.suffix_automaton import build, find_lcs


def random_string(alphabet, length):
    return "".join(random.choices(alphabet, k=length))


def difflib_lcs(s1, s2):
    sm = difflib.SequenceMatcher(
        a=s1,
        b=s2,
        autojunk=False,
    )
    return sm.find_longest_match()


def string_matching_lcs(s1, s2):
    return find_lcs(build(s1), s2)


def timeit(method, method_name, s1, s2, index=False):
    start = time.perf_counter_ns()

    p1, p2, length = method(s1, s2)
    print((p1, p2, length))
    duration_ms = round((time.perf_counter_ns() - start) / 1e6, 0)

    row = {}
    if index:
        row = {
            "Length": f"{(len(s1) + len(s2)) // 1024}k",
            "Match Length": length,
        }
    the_rest = {f"{method_name} (ms)": duration_ms}
    row.update(the_rest)
    return row


difflib_results = []
string_matching_results = []


def make_string_pairs():
    results = []
    for n in range(0, 3, 1):
        length = 1_024 * 2**n
        results.append(tuple(random_string("abc", length) for _ in range(2)))
    return results


def make_table(method, method_name, string_list, index=True):
    results = []

    for strings in string_list:
        results.append(timeit(method, method_name, *strings, index=index))

    return results


def render_table(table):
    return tabulate(
        table,
        headers="keys",
        tablefmt="github",
        intfmt=",d",
        floatfmt=",.0f",
    )


def render_tables():
    string_list = make_string_pairs()
    combined = []
    for r1, r2 in zip(
        make_table(difflib_lcs, "Difflib", string_list, True),
        make_table(string_matching_lcs, "Gaspra", string_list, False),
    ):
        full_row = r1.copy()
        full_row.update(r2)
        combined.append(full_row)

    return render_table(combined)


if __name__ == "__main__":
    print(render_tables())



from itertools import chain

from collections.abc import Sequence

from gaspra.suffix_automaton import (
    build,
    _get_all_start_positions,
)


def concatenate_strings(string_set: Sequence[str | Sequence[int]]):
    # Concatenate the members of string_set with end of string "tokens"
    # to separate them. End of string "tokens" are represented by
    # strings of length 2, e.g. "$0", "$1", etc., so there's no
    # possibility of them matching any single unicode character.

    chainlets = [
        chain(string, (f"${index}",)) for index, string in enumerate(string_set)
    ]

    return chain(*chainlets)


def find_lcs(*string_set: str) -> tuple[tuple[int, ...], int | None]:
    """
    Given a sequence of strings, find the longest substring common
    to all members of the sequence.

    Arguments:
       *string_set: Sequence[str]
          Sequence of strings on which to compute the LCS.

    Returns:
       tuple[int, ...]:
          Tuple with the positions of the first occurence in each string
       length:
          Length of the common substring
    """

    # The suffix automaton can be very deep so we must use an interative
    # solution rather than a recursion solution.  The python recursion
    # limit is on the order of 1k and we want to handle strings in
    # `string_set`` much larger than that.

    # The algorithm works as follows: When we don't know which elements
    # of `string_set` share a given node's state, we add that node to
    # the queue of nodes to be processed.

    # At each iteration, we examine the node on the top of the stack and
    # attempt to compute the strings that share its state.  That is
    # possible if the shared strings are known for all of its children.
    # If any of the chidren have unknown shared string membershiops, add
    # those children to the stack so their shared strings will get
    # computed.

    # Initialize the stack with just the root node, for which string
    # membership is assumed to be initially unknown (the root node is
    # always shared by all of the strings, but setting it initially as
    # unknown forces the algorithm to compute the shared strings for it
    # and all of its descendents.)

    # We're interested in the deepest state that *all* of the strings
    # have in common.  As we're computing the common strings for each
    # state, we'll record the maximal state with *all* strings in
    # common.

    # The values in `shared_strings` are sets that represent the set of
    # strings in `string_set` that share a given node's state.  The key
    # is the node's id.
    shared_strings = {}

    root = build(concatenate_strings(string_set))
    if len(string_set) == 0:
        return (), None

    stack = [root]
    max_length = 0
    best_node = root

    while len(stack) > 0:
        top = stack[-1]
        # If we already know the shared strings for the node on the top
        # of the stack, simply remove it.
        if top.id in shared_strings:
            stack.pop()

        # If we don't know the shared_strings for the top of the stack...
        elif all([node.id in shared_strings for node in top.transitions.values()]):
            update_shared_strings(shared_strings, top)

            if (
                len(shared_strings[top.id]) == len(string_set)
                and top.length > max_length
            ):
                max_length = top.length
                best_node = top

            stack.pop()

        # Or add the unknown children to the stack so they get computed.
        else:
            for next_node in top.transitions.values():
                if next_node.id not in shared_strings:
                    stack.append(next_node)

    raw_positions = _get_all_start_positions(best_node, max_length)

    string_positions = get_string_offsets(raw_positions, string_set)

    return tuple(string_positions), max_length


def get_string_offsets(positions, string_set):
    adjusted_positions = []
    next_string_start_pos = 0
    for s in string_set:
        for position in positions:
            if position >= next_string_start_pos:
                adjusted_positions.append(position - next_string_start_pos)
                break
        next_string_start_pos += len(s) + 1  # + 1 is for the separator.
    return adjusted_positions


def update_shared_strings(shared_strings, top):
    shared_strings[top.id] = set()
    for token, child_node in top.transitions.items():
        # If any of the transitions are end of string tokens, then we
        # know the parent state represents a substring associated with
        #  that token. End of string "tokens" are represented by strings
        # of length 2, e.g. "$0", "$1", etc., so there's no possibility
        # of them matching any single unicode character or any int.
        if not isinstance(token, str) or len(token) > 1:
            shared_strings[top.id].add(token)

        # If any of the transitions are regular charactrers then the
        # parent inherits the child's memberships
        else:
            child_memberships = shared_strings[child_node.id]
            shared_strings[top.id] = shared_strings[top.id].union(child_memberships)



from __future__ import annotations

from collections.abc import Iterator, Iterable, Sequence
from dataclasses import dataclass, field
import random
from typing import Callable

from gaspra.types import TokenSequence


@dataclass
class Node:
    id: int
    length: int = 0
    transitions: dict[str | int, Node] = field(default_factory=dict)
    first_endpos: int = 0  # First possible ending position of substring.
    is_terminal: bool = False  # Is this a terminal state?
    link: Node | None = None
    reverse_links: list[Node] = field(default_factory=list)
    empty: str | TokenSequence = ""

    @staticmethod
    def get_factory(empty_sequence: str | TokenSequence):
        generator = make_id_sequence()

        def factory(**kwargs):
            return Node(id=next(generator), empty=empty_sequence, **kwargs)

        return factory


def make_id_sequence() -> Iterator[int]:
    id = 0
    while True:
        yield id
        id += 1


def wrap_node_factory(
    node_factory: Callable[..., Node],
    node_list: list[Node],
):
    """Wrap a node_factory() so that it always appends to a node_list

    Arguments:
        node_factory: Callable[..., Node]
            Node factory to wrap
        node_list: List[Node]
            List to append to when a node gets created
    """

    def wrapped_node_factory(**kwargs):
        new_node = node_factory(**kwargs)
        node_list.append(new_node)
        return new_node

    return wrapped_node_factory


def build(
    input_string: Iterable[str | int],
    reverse_links=True,
    mark_terminals=True,
    empty: str | tuple[int, ...] = "",
) -> Node:
    """Build a suffix automaton from `input_string`.

    Arguments:
        input_string:  Iterable[str | int]
            Character or byte sequence representing the input.
        reverse_links: bool
            If true, construct reverse_links (needed to find *all*
            occurrences of a string.
        mark_terminals: bool
            If true, flag terminal nodes as such (needed to read suffixes
            back from automaton).  This is typically very fast.

    """
    # Create a list to capture all created nodes to make it easier to
    # iterate over them later.
    node_list = []

    # Decorate node factory so that node_list is updated on every call.
    node_factory = wrap_node_factory(Node.get_factory(empty), node_list)

    # Implemention mostly follows description at
    # https://cp-algorithms.com/string/suffix-automaton.html

    root = node_factory()
    current = root

    for character in input_string:
        current = extend(character, current, node_factory)

    if mark_terminals:
        mark_terminal_nodes(current)

    if reverse_links:
        add_reverse_links(node_list)

    return root


def extend(
    character: str | int,
    last: Node,
    node_factory: Callable[..., Node],
):
    """Extend a partially constructed automaton by one additional character

    Given a suffix automaton that recognizes the suffixes of some string
    "s", extend that automoton to recognize the suffixes of `s +
    character`

    """
    new_node = node_factory(
        length=last.length + 1,
        first_endpos=last.length + 1,
    )

    previous = None
    current = last
    while current is not None and character not in current.transitions:
        current.transitions[character] = new_node
        previous = current
        current = current.link

    if current is None:
        new_node.link = previous
    else:
        q = current.transitions[character]
        if q.length == current.length + 1:
            new_node.link = q
        else:
            new_node.link = insert_node(character, current, q, node_factory)

    return new_node


def insert_node(
    character: str | int,
    p: Node,
    q: Node,
    node_factory: Callable[..., Node],
):
    clone = node_factory(
        length=p.length + 1,
        first_endpos=q.first_endpos,
        link=q.link,
        transitions=q.transitions.copy(),
    )
    q.link = clone

    current = p
    while (
        current is not None
        and character in current.transitions
        and current.transitions[character] == q
    ):
        current.transitions[character] = clone
        current = current.link

    return clone


def mark_terminal_nodes(final_node: Node):
    current = final_node
    while current is not None:
        current.is_terminal = True
        current = current.link
    return


def add_reverse_links(node_list: list[Node]):
    for node in node_list:
        if node.link is not None:
            node.link.reverse_links.append(node)
    return


def find_substring(root: Node, s: Sequence[str | int]) -> int | None:
    current = _find_match_node(root, s)
    if current is None:
        return None
    return current.first_endpos - len(s)


def find_substring_all(root: Node, s: str) -> Iterable[int]:
    current = _find_match_node(root, s)
    if current is None:
        return ()

    return _get_all_start_positions(current, len(s))


def _get_all_start_positions(current, length):
    candidates = sorted(
        node.first_endpos - length for node in follow_reverse_links(current)
    )

    positions = dedup_sorted(candidates)
    return positions


def _find_match_node(root: Node, s: Iterable[str | int]) -> Node | None:
    current = root
    for character in s:
        current = current.transitions.get(character)
        if current is None:
            return None
    return current


def follow_reverse_links(current):
    # Use a non-resursive solution
    to_do = [current]
    all = []
    while len(to_do) > 0:
        current = to_do.pop()
        all.append(current)
        for child in current.reverse_links:
            to_do.append(child)
    return all


def dedup_sorted(sequence: Iterable[int]) -> list[int]:
    deduped = []
    for candidate in sequence:
        if len(deduped) == 0 or candidate != deduped[-1]:
            deduped.append(candidate)
    return deduped


def find_lcs(root: Node, s: str | TokenSequence) -> tuple[int, int, int]:
    """Return the starting positions and length of an LCS.

    The LCS is the longest common substring of the
    string used to construct the automaton rooted at `root`
    and the passed string `s`.

    Arguments:
        root: Node
            Root node of a suffix automaton for one of the strings
        s: str
            The string searched for substrings in common with the
            automaton string

    Returns:
        int
            The starting position of the match in the automaton string
        int
            The starting position of the match in `s`
        int
            The length of the common substring

    """
    longest_match_length = current_match_length = 0
    longest_match_s1_endpos = longest_match_s2_endpos = 0

    current = root
    for position, character in enumerate(s):
        while (
            next := current.transitions.get(character)
        ) is None and current.link is not None:
            current = current.link
            current_match_length = current.length

        if next is not None:
            current_match_length += 1
            current = next
        # This block can be indented but the sample implementation has
        # it this way.
        if current_match_length > longest_match_length:
            longest_match_length = current_match_length
            longest_match_s1_endpos = current.first_endpos
            longest_match_s2_endpos = position + 1

    return (
        longest_match_s1_endpos - longest_match_length,
        longest_match_s2_endpos - longest_match_length,
        longest_match_length,
    )


def all_suffixes(current: Node) -> Iterator[str | tuple[int, ...]]:
    """Iterate over every suffix in the automaton.

    The only purpose for this is in a test that ensures the automaton
    produces all suffixes and only suffixes.
    """
    if current.is_terminal:
        yield current.empty

    for token, node in current.transitions.items():
        for substring in all_suffixes(node):
            if isinstance(token, str) and isinstance(substring, str):
                yield token + substring
            elif isinstance(token, int) and isinstance(substring, tuple):
                yield (token,) + substring  # (character,)  # + substring
            else:
                raise ValueError(f"Unexpected type {type(token)}")


def dump(root: Node, indent: int = 0, dumped: set[int] | None = None):
    if dumped is None:
        dumped = set()

    if root.id in dumped:
        return

    print(
        f"{' '*indent}{root.id}: (length:{root.length}, "
        f"is_terminal:{root.is_terminal})"
    )
    if root.transitions.items():
        for character, node in root.transitions.items():
            print(f"{' '*indent}  {character} -> {node.id}")
    else:
        print(f"{' '*indent}  No transitions")

    dumped.add(root.id)

    for character, node in root.transitions.items():
        dump(node, indent, dumped)


if __name__ == "__main__":
    random_string = "".join(random.choices("XYZ", k=10))
    for string in ("abcbc", "bananas", random_string):
        print(f"\n{string}")
        root = build(string)
        dump(root, indent=2)

        print("\nAll suffixes:")
        for item in sorted(all_suffixes(root), key=len):
            print(f"'{item}'")
        print()

        def show_match(substring):
            position = find_substring(root, substring)
            print(f"'{substring}': {'no' if position is None else 'yes'} ({position})")

        for k in range(len(string)):
            suffix = string[k:]
            prefix = string[:-k]
            random_substring = "".join(random.choices(string, k=len(string) - k))

            show_match(suffix)
            show_match(prefix)
            show_match(random_substring)



from gaspra.types import TokenSequence


from collections.abc import Iterable, Sequence


def line_tokenize(
    *string_set: str,
):
    tokenized = []
    token_dict = {}

    for s in string_set:
        lines = s.split("\n")
        # Ignore the empty string that gets generated
        # by an ending newline.

        if len(lines) > 0 and lines[-1] == "":
            lines = lines[:-1]

        for line in s.split("\n"):
            if line not in token_dict:
                token_dict[line] = len(token_dict)

        tokenized.append(tuple(token_dict[line] for line in lines))
    token_map = tuple(token_dict.keys())
    return tuple([token_map, *tokenized])



import io
from gaspra.suffix_automaton import build, Node


def dump(node: Node, f: io.TextIOWrapper):
    queue = [node]
    processed = set()

    while queue:
        current = queue.pop()

        if current.id not in processed:
            if current.is_terminal:
                f.write(f" {current.id}[color=red]\n")

            for token, child in current.transitions.items():
                f.write(f' {current.id} -> {child.id} [label="{token}"]\n')
                queue.append(child)

        processed.add(current.id)

    return


def dot_dump(s: str, filename: str):
    root = build(s)

    with open(filename, "w") as f:
        f.write("digraph G {\n")
        f.write("  rankdir=LR\n")
        dump(root, f)
        f.write("}\n")


if __name__ == "__main__":
    dot_dump("abc", "abc.dot")
    dot_dump("ababab", "ababab.dot")
    dot_dump("abc0cdabchi1abcabcxyz", "multiple.dot")



from contextlib import contextmanager
from copy import deepcopy
import io
from rich.console import Console

from gaspra.types import Change


def rich_escape(s):
    return s.replace("[", r"\[")


SCREEN_MARKUP = {
    "fragment": {
        "into": {"prefix": "[bright_green]", "suffix": "[/]"},
        "from": {"prefix": "[bright_red]", "suffix": "[/]"},
    },
    "line": {
        "into": {"prefix": lambda _: "[green]", "suffix": lambda _: "[/]"},
        "from": {"prefix": lambda _: "[red]", "suffix": lambda _: "[/]"},
    },
    "escape": rich_escape,
    "separator": "",
    "header": {"prefix": "<<<[bright_blue]", "suffix": "[/]>>>\n"},
}

STRIKEOUT_SCREEN_MARKUP = deepcopy(SCREEN_MARKUP)
STRIKEOUT_SCREEN_MARKUP["fragment"]["from"]["prefix"] = "[bright_red strike]"
STRIKEOUT_SCREEN_MARKUP["line"]["from"]["prefix"] = "[red strike]"

GIT_MARKUP = {
    "fragment": {
        "into": {"prefix": "", "suffix": ""},
        "from": {"prefix": "", "suffix": ""},
    },
    "line": {
        "into": {"prefix": lambda s: f"<<<<<<< {s}\n", "suffix": lambda _: ""},
        "from": {"prefix": lambda _: "", "suffix": lambda s: f">>>>>>> {s}\n"},
    },
    "escape": lambda _: _,
    "separator": "=======\n",
    "header": {"prefix": "<<<", "suffix": ">>>\n"},
}

TOKEN_GIT_MARKUP = {
    "into": {"prefix": lambda s: f"<<<<<<< {s}\n", "suffix": lambda _: ""},
    "from": {"prefix": lambda _: "", "suffix": lambda s: f">>>>>>> {s}\n"},
    "escape": lambda _: _,
    "separator": "=======\n",
    "header": {"prefix": "<<<", "suffix": ">>>"},
}

TOKEN_SCREEN_MARKUP = {
    "into": {"prefix": lambda _: "[bright_green]", "suffix": lambda _: "[/]"},
    "from": {"prefix": lambda _: "[bright_red]", "suffix": lambda _: "[/]"},
    "escape": rich_escape,
    "separator": "",
    "header": {"prefix": "<<<", "suffix": ">>>"},
}


def show_header(print, header, markup={}):
    escape = markup.get("escape", lambda _: _)
    header_markup = markup.get("header", {})
    prefix = header_markup.get("prefix", "")
    suffix = header_markup.get("suffix", "")
    if header:
        print(f"\n{prefix}{escape(header)}{suffix}")


def markup_changes(
    print,
    fragment_sequence,
    __name0__,
    __name1__,
    markup={},
    header: str | None = None,
    **__kwargs__,
):
    escape = markup.get("escape", lambda _: _)

    show_header(print, header, markup)

    fragment_markup = markup.get("fragment", {})

    def print_fragment(fragment, fragment_markup):
        prefix = fragment_markup["prefix"]
        suffix = fragment_markup["suffix"]
        print(f"{prefix}{escape(fragment)}{suffix}")

    for fragment in fragment_sequence:
        if isinstance(fragment, str):
            print(fragment)

        if isinstance(fragment, tuple):
            print_fragment(fragment[0], fragment_markup["into"])
            print_fragment(fragment[1], fragment_markup["from"])


def print_line(print, line, name, line_markup):
    prefix = line_markup["prefix"](name)
    suffix = line_markup["suffix"](name)
    print(f"{prefix}{line}{suffix}")


def _markup_and_add_fragment(partial_line, fragment, fragment_markup):
    if fragment:
        prefix = fragment_markup["prefix"]
        suffix = fragment_markup["suffix"]
        return partial_line + f"{prefix}{fragment}{suffix}"
    else:
        return partial_line


def markup_and_add_fragment(
    partial_line_into, partial_line_from, fragment, fragment_markup
):
    partial_line_into = _markup_and_add_fragment(
        partial_line_into, fragment[0], fragment_markup["into"]
    )
    partial_line_from = _markup_and_add_fragment(
        partial_line_from, fragment[1], fragment_markup["from"]
    )
    return partial_line_into, partial_line_from


def update_partials(partial_line_into, partial_line_from, input_fragment):
    if (partial_line_into and partial_line_into[-1] != "\n") or (
        partial_line_from and partial_line_from[-1] != "\n"
    ):
        partial_line_into = partial_line_into + input_fragment
        partial_line_from = partial_line_from + input_fragment
        input_fragment = ""
    return partial_line_into, partial_line_from, input_fragment


def conflict_finisher(print, markup, name0, name1):
    line_markup = markup.get("line", {})
    escape = markup.get("escape", lambda _: _)

    def finish_conflict(partial_line_into, partial_line_from, input_fragment):
        input_fragment = escape(input_fragment)
        partial_line_into, partial_line_from, input_fragment = update_partials(
            partial_line_into, partial_line_from, input_fragment
        )

        print_line(
            print,
            partial_line_into,
            name0,
            line_markup["into"],
        )
        print(markup["separator"])
        print_line(
            print,
            partial_line_from,
            name1,
            line_markup["from"],
        )
        print(input_fragment)

    return finish_conflict


def line_oriented_markup_changes(
    print,
    fragment_sequence,
    name0,
    name1,
    markup={},
    header: str | None = None,
    **__kwargs__,
):
    show_header(print, header, markup)
    escape = markup.get("escape", lambda _: _)
    fragment_markup = markup.get("fragment", {})

    finish_conflict = conflict_finisher(print, markup, name0, name1)

    in_conflict = False
    partial_line_into = partial_line_from = ""
    for fragment in fragment_sequence:
        if isinstance(fragment, str):
            lines = fragment.split("\n")
            if len(lines) > 1:  # Have a newline
                if in_conflict:
                    finish_conflict(
                        partial_line_into,
                        partial_line_from,
                        lines[0] + "\n",
                    )
                    print(escape(join(lines[1:-1])))
                else:
                    print(escape(join(lines[:-1])))
                in_conflict = False
                partial_line_into = partial_line_from = lines[-1]
            else:
                partial_line_into += lines[0]
                partial_line_from += lines[0]

        if isinstance(fragment, tuple):
            in_conflict = True
            partial_line_into, partial_line_from = markup_and_add_fragment(
                partial_line_into, partial_line_from, fragment, fragment_markup
            )
    if in_conflict:
        if partial_line_into[-1:] != "\n" or partial_line_from[-1:] != "\n":
            tail = "\n"
        else:
            tail = ""
        finish_conflict(partial_line_into, partial_line_from, tail)
    elif partial_line_from:
        # If not in a conflict, partial_line_into should be
        # exactly the same as partial_line_from.
        print(escape(partial_line_from) + "\n")


def join(lines):
    if len(lines) > 0:
        return "\n".join(line for line in lines) + "\n"
    else:
        return ""


@contextmanager
def file_writer(filename):
    file = open(filename, "wt", encoding="utf-8")

    yield file.write
    file.close()


@contextmanager
def console_writer():
    console = Console(force_terminal=True, highlight=False)

    def print(s):
        console.print(s, end="")

    yield print
    return


def print_conflict(print, version, token_map, escape, name, markup):
    prefix = markup["prefix"](name)
    suffix = markup["suffix"](name)
    # Collect output in a buffer so that prefix and suffix
    # tags appear in the same call to print().
    # That is important for the rich Console writer.
    buffer = io.StringIO()
    if prefix is not None:
        buffer.write(prefix)
    for token in version:
        buffer.write(escape(token_map[token]))
        buffer.write("\n")
    if suffix is not None:
        buffer.write(suffix)
    print(buffer.getvalue())


def token_oriented_markup_changes(
    print,
    fragment_sequence,
    name0,
    name1,
    markup={},
    header: str | None = None,
    token_map=(),
):
    escape = markup["escape"]

    if header:
        show_header(print, header, markup)

    for item in fragment_sequence:
        if isinstance(item, Change):
            print_conflict(print, item.a, token_map, escape, name0, markup["into"])
            print(markup["separator"])
            print_conflict(print, item.b, token_map, escape, name1, markup["from"])
        else:
            for token in item:
                print(token_map[token])
                print("\n")



from __future__ import annotations
from collections.abc import Iterable
from dataclasses import dataclass
from itertools import chain
import os

from gaspra.common import DATA_DIR
from gaspra.suffix_automaton import build, find_lcs
from gaspra.types import Change, ChangeIterable, TokenSequence


@dataclass
class CopyFragment:
    insert: str | TokenSequence
    length: int


@dataclass
class ChangeFragment:
    insert: str | TokenSequence
    delete: str | TokenSequence
    length: int


@dataclass
class ConflictFragment:
    version1: str | TokenSequence
    version2: str | TokenSequence


@dataclass
class ChangesetLeaf:
    original: str | TokenSequence
    modified: str | TokenSequence

    original_slice: slice
    modified_slice: slice

    def changes(self):
        yield self

    def _fragments(
        self, _: str | TokenSequence
    ) -> Iterable[ChangeFragment | CopyFragment]:
        # Construction of the tree creates "empty" changesets.
        # Omit those from the output stream.
        if self.modified or self.original:
            yield ChangeFragment(
                insert=self.modified,
                delete=self.original,
                length=len(
                    self.original,
                ),
            )

    def fragments(self, _: str | TokenSequence) -> ChangeIterable:
        # Construction of the tree creates "empty" changesets.
        # Omit those from the output stream.
        if self.modified or self.original:
            yield Change(self.modified, self.original)

    def apply_forward(self, _: str | TokenSequence):
        yield self.modified

    def apply_reverse(self, _: str | TokenSequence):
        yield self.original

    def show(
        self, __original__: str | TokenSequence, __modified__: str | TokenSequence
    ) -> str:
        result = ""

        if self.original:
            result += f"[red strike]{escape(self.original)}[/]"
            # result += f"[delete]{escape(self.original_str)}[/delete]"
        if self.modified:
            result += f"[green]{escape(self.modified)}[/]"
        # result += f"[insert]{escape(self.modified_str)}[/insert]"

        return result

    def __str__(self):
        result = ""
        if self.original:
            result += f"original: {self.original}\n"
        if self.modified:
            result += f"modified: {self.modified}\n"
        return result


@dataclass
class Changeset:
    common_original: slice
    common_modified: slice

    prefix: Changeset | ChangesetLeaf
    suffix: Changeset | ChangesetLeaf

    def changes(self):
        yield from self.prefix.changes()
        yield self
        yield from self.suffix.changes()

    def _fragments(
        self, original: str | TokenSequence
    ) -> Iterable[ChangeFragment | CopyFragment]:
        yield from self.prefix._fragments(original)
        copy = original[self.common_original]
        yield CopyFragment(insert=copy, length=len(copy))
        yield from self.suffix._fragments(original)

    def fragments(self, original: str | TokenSequence) -> ChangeIterable:
        yield from self.prefix.fragments(original)
        yield original[self.common_original]
        yield from self.suffix.fragments(original)

    def apply_forward(self, original: str | TokenSequence):
        yield from self.prefix.apply_forward(original)
        yield original[self.common_original]
        yield from self.suffix.apply_forward(original)

    def apply_reverse(self, modified: str | TokenSequence):
        yield from self.prefix.apply_reverse(modified)
        yield modified[self.common_modified]
        yield from self.suffix.apply_reverse(modified)

    def show(self, original: str, modified: str | TokenSequence):
        return (
            self.prefix.show(original, modified)
            + escape(original[self.common_original])
            + self.suffix.show(original, modified)
        )

    def __str__(self):
        s_original = f"{self.common_original.start}:{self.common_original.stop}"
        s_modified = f"{self.common_modified.start}:{self.common_modified.stop}"
        return f"original[{s_original}]/modified[{s_modified}]\n"


def diff(
    original: str | TokenSequence, modified: str | TokenSequence
) -> ChangeIterable:
    """Returns the changes between a and b.

    Arguments:
        original: str
            The "original" tring
        modified: str
            The "modified" string

    Returns:
        Iterable[str]
            The changes between a and b.  Each item in the sequence
            is either a string or a tuple of two strings.  If the item
            is a string, it is unchanged test between `original` and
            ``modified`.  If the item is a tuple, the first string is the
            string inserted at that point in `original` to get `modified`
            and the second string is the string that was deleted.

    """
    changeset = find_changeset(original, modified)
    yield from changeset.fragments(original)


def find_changeset(
    original: str | TokenSequence,
    modified: str | TokenSequence,
    original_slice: slice = slice(0, None),
    modified_slice: slice = slice(0, None),
) -> Changeset | ChangesetLeaf:
    automaton_original = build(original[original_slice])

    common_offset_original, common_offset_modified, common_length = find_lcs(
        automaton_original, modified[modified_slice]
    )

    if common_length == 0:
        changeset = ChangesetLeaf(
            original=original[original_slice],
            modified=modified[modified_slice],
            original_slice=original_slice,
            modified_slice=modified_slice,
        )
    else:
        common_original = slice(
            original_slice.start + common_offset_original,
            original_slice.start + common_offset_original + common_length,
        )
        common_modified = slice(
            modified_slice.start + common_offset_modified,
            modified_slice.start + common_offset_modified + common_length,
        )

        prefix = find_changeset(
            original,
            modified,
            slice(original_slice.start, common_original.start),
            slice(modified_slice.start, common_modified.start),
        )
        suffix = find_changeset(
            original,
            modified,
            slice(common_original.stop, original_slice.stop),
            slice(common_modified.stop, modified_slice.stop),
        )
        changeset = Changeset(
            prefix=prefix,
            suffix=suffix,
            common_original=common_original,
            common_modified=common_modified,
        )

    return changeset


def apply_forward(changeset, original: str | Iterable[int]):
    changed = changeset.apply_forward(original)
    if isinstance(original, str):
        patched_original = "".join(changed)
    else:
        patched_original = tuple(chain(*changed))

    return patched_original


def apply_reverse(changeset, modified: str):
    changed = changeset.apply_reverse(modified)
    if isinstance(modified, str):
        reverse_patched_modified = "".join(changed)
    else:
        reverse_patched_modified = tuple(chain(*changed))

    return reverse_patched_modified


def escape(s):
    return s.replace("[", r"\[")
    # return s


if __name__ == "__main__":
    from rich.console import Console

    console = Console(highlight=False)

    with open(os.path.join(DATA_DIR, "file1")) as file1, open(
        os.path.join(DATA_DIR, "file2")
    ) as file2:
        original = file1.read()
        modified = file2.read()

    changeset = find_changeset(original, modified)

    markup = changeset.show(original, modified)

    patched_original = apply_forward(changeset, original)
    reverse_patched_modified = apply_reverse(changeset, modified)

    assert patched_original == modified
    assert reverse_patched_modified == original

    console.print(markup)



from dataclasses import replace

from gaspra.changesets import (
    ChangeFragment,
    ConflictFragment,
    CopyFragment,
    find_changeset,
)
from gaspra.common import get_joiner, common_prefix_length
from gaspra.types import Change, TokenSequence


OutputType = CopyFragment | ChangeFragment | ConflictFragment
InputType = CopyFragment | ChangeFragment


def merge(
    parent: str | TokenSequence,
    branch0: str | TokenSequence,
    branch1: str | TokenSequence,
):
    empty = parent[0:0]

    changeset0 = find_changeset(parent, branch0)
    changeset1 = find_changeset(parent, branch1)

    fragments0 = list(reversed(list(changeset0._fragments(parent))))
    fragments1 = list(reversed(list(changeset1._fragments(parent))))

    merged = _merge(fragments0, fragments1)

    return consolidate(merged, empty=empty)


def consolidate(fragments, empty: str | TokenSequence = ""):
    """Consolidate consecutive fragments of similar types
    into a single segment"""

    # Make two passes though the output.  In the first
    # past, conflicts and consolidated (and possibly expanded
    # into copy and conflict fragments
    # In the second pass, consolidate copy and conflict fragments
    # with one another and don't expand the conflict fragments.
    # The second pass is required because some of the copy fragments
    # that get consolidated in the second pass are created from
    # the conflict expansions in the first pass.

    yield from consolidate_all(
        consolidate_conflicts(fragments, empty=empty), empty=empty
    )

    return


def consolidate_conflicts(input_stack, empty: str | TokenSequence = ""):
    join = get_joiner(empty)
    input_stack = list(reversed(list(input_stack)))

    while input_stack:
        # Loop through and consolidate contiguous ConflictFragments
        # First collect a contiguous gruop
        conflict_group = []
        while input_stack and isinstance(input_stack[-1], ConflictFragment):
            conflict_group.append(input_stack.pop())

        # Do a combination of consolidation and re-splitting with find_changeset()
        version1 = join(fragment.version1 for fragment in conflict_group)
        version2 = join(fragment.version2 for fragment in conflict_group)

        for fragment in find_changeset(version2, version1)._fragments(version2):
            if isinstance(fragment, CopyFragment):
                yield fragment
            elif isinstance(fragment, ChangeFragment):
                yield ConflictFragment(
                    fragment.insert,
                    fragment.delete,
                )
            else:
                raise ValueError(f"Unexpected fragment type: {type(fragment)}")

        # For now, just copy the other fragment types.  They will be
        # consolidated in the second pass.
        while input_stack and isinstance(
            input_stack[-1], CopyFragment | ChangeFragment
        ):
            yield input_stack.pop()


def consolidate_all(staged, empty: str | TokenSequence = ""):
    join = get_joiner(empty)

    staged = list(reversed(list(staged)))

    something_has_been_output = False
    while staged:
        # Loop through and consolidate contiguous ConflictFragments
        conflict_group = []
        while staged and isinstance(staged[-1], ConflictFragment):
            conflict_group.append(staged.pop())
        version1 = join(fragment.version1 for fragment in conflict_group)
        version2 = join(fragment.version2 for fragment in conflict_group)

        if version1 or version2:
            yield Change(version1, version2)
            something_has_been_output = True

        # Loop through and consolidate the other types
        copy_or_change_group = []
        while staged and isinstance(staged[-1], CopyFragment | ChangeFragment):
            copy_or_change_group.append(staged.pop())
        insert = join(fragment.insert for fragment in copy_or_change_group)
        if insert:
            yield insert
            something_has_been_output = True

    if not something_has_been_output:
        yield empty


def _merge(fragments0: list[InputType], fragments1):
    """Process changes in fragments0 and fragments1 until
    they are empty.  Note, this function modifies the passed lists"""

    within_conflict = False

    while fragments0 and fragments1:
        fragment0 = fragments0.pop()
        fragment1 = fragments1.pop()

        output, tail0, tail1, within_conflict = process_fragments(
            fragment0,
            fragment1,
            within_conflict,
        )

        if output:
            yield output

        # If the fragments weren't fully processed, push their tails
        # back onto their respective stacks.

        if tail0:
            fragments0.append(tail0)

        if tail1:
            fragments1.append(tail1)

    # We've exhausted one of the input queues.  Flush the
    # other one if there's anything left.
    if fragments0 or fragments1:
        yield from flush_remaining(fragments0, fragments1, within_conflict)


def flush_remaining(fragments0, fragments1, within_conflict):
    remaining = fragments0 or fragments1
    for item in reversed(remaining):
        if within_conflict and remaining == fragments0:
            yield ConflictFragment(item.insert, "")
        elif within_conflict:
            yield ConflictFragment("", item.insert)
        else:
            yield item


def process_fragments(fragment0, fragment1, within_conflict: bool):
    if isinstance(fragment0, CopyFragment) and isinstance(fragment1, CopyFragment):
        output, tail0, tail1 = copy_copy(fragment0, fragment1)
        within_conflict = False

    elif within_conflict:
        output, tail0, tail1 = pending_conflict(fragment0, fragment1)

    elif isinstance(fragment0, ChangeFragment) and isinstance(
        fragment1, ChangeFragment
    ):
        output, tail0, tail1, within_conflict = change_change(fragment0, fragment1)

    elif isinstance(fragment0, CopyFragment) and isinstance(fragment1, ChangeFragment):
        output, tail0, tail1 = copy_change(fragment0, fragment1)

    elif isinstance(fragment0, ChangeFragment) and isinstance(fragment1, CopyFragment):
        output, tail1, tail0 = copy_change(fragment1, fragment0)

    else:
        raise RuntimeError(f"Unexpected types: {type(fragment0)} or {type(fragment1)}")

    return output, tail0, tail1, within_conflict


def pending_conflict(fragment0: InputType, fragment1: InputType):
    min_length = min(fragment0.length, fragment1.length)

    head0, tail0 = split_fragment(fragment0, min_length)
    head1, tail1 = split_fragment(fragment1, min_length)

    from0 = head0.insert if head0 else ""
    from1 = head1.insert if head1 else ""

    output = ConflictFragment(from0, from1)
    return output, tail0, tail1


def copy_copy(fragment0: CopyFragment, fragment1: CopyFragment):
    if fragment0.length < fragment1.length:
        shorter, longer = fragment0, fragment1
    else:
        shorter, longer = fragment1, fragment0

    long_queue_tail = None
    if shorter.length != longer.length:
        *_, long_queue_tail = split_copy_fragment(longer, shorter.length)

    tail0 = tail1 = None

    if fragment0 == longer:
        tail0 = long_queue_tail
    else:
        tail1 = long_queue_tail

    return shorter, tail0, tail1


def copy_change(copy_fragment, change_fragment):
    output = copy_tail = change_tail = None

    smaller_length = min(copy_fragment.length, change_fragment.length)
    if change_fragment.length == smaller_length:
        output = change_fragment

        # Anything left over?
        if copy_fragment.length > smaller_length:
            *_, copy_tail = split_copy_fragment(copy_fragment, smaller_length)

    else:  # Change is longer than copy implies a conflict.
        head0, head1 = split_change_fragment(
            change_fragment, len(change_fragment.insert), smaller_length
        )
        if head0:
            output = ConflictFragment(head0.insert, copy_fragment.insert)

        if head1:
            change_tail = head1

    return output, copy_tail, change_tail


def change_change(fragment0: ChangeFragment, fragment1: ChangeFragment):
    """Handle two change blocks appearing at the same location"""

    if has_composable_changes(fragment0, fragment1):
        return (*compose_changes(fragment0, fragment1), False)

    # Determine how much is common between the two changesets
    insert_length, delete_length = common_prefix_lengths(fragment0, fragment1)

    # Exactly the same changeset can simply be passed along.
    if are_identical_changes(fragment0, fragment1, insert_length, delete_length):
        return fragment0, None, None, False

    return (*ordinary_conflict(fragment0, fragment1), True)


def has_composable_changes(fragment0, fragment1):
    # This is an edge case that showed up in testing. If one change is a
    # pure insertion (no deletion) and the other is a pure deletion (no
    # insertion), they can be combined into a single conflict-free
    # change.  It needs to be pushed back onto the input list to be
    # processed properly One input sequence was [insert x/delete
    # nothing][s] The other input sequence was [insert nothing/delete
    # s]. These are transformed to the sequences 1) [s] and 2) [insert
    # x/delete s] by pushing [insert x/delete s] back onto the input
    # queue which has the affect of inserting 's' at position where 's'
    # was.  See compose_changes() for how the changes are composed.

    return (fragment0.length == 0 and len(fragment1.insert) == 0) or (
        len(fragment0.insert) == 0 and fragment1.length == 0
    )


def compose_changes(fragment0, fragment1):
    if fragment0.length == 0 and len(fragment1.insert) == 0:
        tail1 = ChangeFragment(
            fragment0.insert,
            fragment1.delete,
            fragment1.length,
        )
        return None, None, tail1
    elif len(fragment0.insert) == 0 and fragment1.length == 0:
        tail0 = ChangeFragment(
            fragment1.insert,
            fragment0.delete,
            fragment0.length,
        )
        return None, tail0, None
    else:
        raise ValueError("Changes are not composable")


def are_identical_changes(
    fragment0,
    fragment1,
    insert_length,
    delete_length,
):
    return (len(fragment0.insert) == len(fragment1.insert) == insert_length) and (
        fragment0.length == fragment1.length == delete_length
    )


def ordinary_conflict(fragment0, fragment1):
    output = None

    length = min(fragment0.length, fragment1.length)
    head0, tail0 = split_change_fragment(
        fragment0,
        len(fragment0.insert),
        length,
    )
    head1, tail1 = split_change_fragment(
        fragment1,
        len(fragment1.insert),
        length,
    )

    if head0 and head1:
        output = ConflictFragment(head0.insert, head1.insert)

    return output, tail0, tail1


def split_fragment(fragment: InputType, length: int):
    if isinstance(fragment, CopyFragment):
        return split_copy_fragment(fragment, length)
    if isinstance(fragment, ChangeFragment):
        return split_change_fragment(fragment, len(fragment.insert), length)


def split_copy_fragment(fragment: CopyFragment, length: int):
    head = tail = None
    if length > 0:
        head = replace(
            fragment,
            insert=fragment.insert[:length],
            length=length,
        )
    if length < fragment.length:
        tail = replace(
            fragment,
            insert=fragment.insert[length:],
            length=fragment.length - length,
        )
    return head, tail


def split_change_fragment(
    fragment: ChangeFragment,
    insert_length,
    length: int,
):
    head = tail = None
    if length > 0 or insert_length > 0:
        head = replace(
            fragment,
            insert=fragment.insert[:insert_length],
            delete=fragment.delete[:length],
            length=length,
        )
    if length < fragment.length or insert_length < len(fragment.insert):
        tail = replace(
            fragment,
            insert=fragment.insert[insert_length:],
            delete=fragment.delete[length:],
            length=fragment.length - length,
        )
    return head, tail


def common_prefix_lengths(
    change1: ChangeFragment,
    change2: ChangeFragment,
):
    # Note: don't use the commonprefix() function from
    # os.path because this one has to work on sequences of ints

    insert_length = common_prefix_length(change1.insert, change2.insert)
    delete_length = common_prefix_length(change1.delete, change2.delete)

    return insert_length, delete_length



from collections.abc import Iterable
from itertools import chain
from typing import Callable
import os

from gaspra.types import StringIterable, TokenSequenceIterable, TokenSequence

DATA_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "..", "data"),
)


def string_joiner(g: StringIterable) -> str:
    return "".join(g)


def tuple_joiner(g: TokenSequenceIterable) -> TokenSequence:
    return tuple(chain(*g))


StringJoiner = Callable[[Iterable[str]], str]
TokenJoiner = Callable[[TokenSequenceIterable], TokenSequence]


def get_joiner(empty) -> StringJoiner | TokenJoiner:
    if isinstance(empty, str):
        joiner = string_joiner
    else:
        joiner = tuple_joiner
    return joiner


def common_prefix_length(a: str | TokenSequence, b: str | TokenSequence):
    length = min(len(a), len(b))

    for length, (x, y) in enumerate(zip(a, b)):
        if x != y:
            return length

    return length


if __name__ == "__main__":
    print(DATA_DIR)



from collections.abc import Sequence
import random


def random_string(alphabet: str, length, seed=42):
    rng = random.Random(seed)
    return "".join(rng.choices(alphabet, k=length))


def random_tokens(alphabet: Sequence[int], length, seed=42):
    rng = random.Random(seed)
    return tuple(rng.choices(alphabet, k=length))


def tokenize(string_or_fragment_list):
    if isinstance(string_or_fragment_list, str):
        return tuple(ord(character) for character in string_or_fragment_list)

    elif isinstance(string_or_fragment_list, tuple | list):
        return tuple(tokenize(fragment) for fragment in string_or_fragment_list)
    else:
        raise TypeError(f"Unsupported fragment type: {type(string_or_fragment_list)}")



import gaspra
import gaspra.changesets


def merge_example():
    original = "The quick brown fox jumps over the lazy dog near the riverbank."
    branch1 = "The quick brown fox leaps over the lazy dogs near the river."
    branch2 = "The quick, clever fox jumps across the lazy dogs by the riverbank."
    conflicts_with_1 = (
        "The swift, agile fox leaps over the sleepy dog near the riverside."
    )

    print(f"Original:\n   {original}")
    print(f"Editor1:\n   {branch1}")
    print(f"Changes:\n   {list(gaspra.changesets.diff(original, branch1))}\n")

    print(f"Editor2:\n   {branch2}")
    print(f"Changes:\n   {list(gaspra.changesets.diff(original, branch2))}")
    print(f"Merge:   {list(gaspra.merge(original, branch1, branch2))}\n")

    print(f"Editor2 (alt):\n   {conflicts_with_1}")
    print(f"Changes:\n   {list(gaspra.changesets.diff(original, conflicts_with_1))}")
    print(f"Merge:   {list(gaspra.merge(original, branch1, conflicts_with_1))}\n")


if __name__ == "__main__":
    merge_example()



import os

TEST_CASE_DIRECTORY = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__),
        "..",
        "..",
        "test-cases",
    )
)
PROGRAM_NAME = os.path.basename(__file__)


TEST_CASES = [
    "About.java",
    "ErrorMsg.java",
    "FlownetController.cpp",
    "Misc.java",
    "RearViewMirror.java",
    "ServerWin_2005.vcproj",  # This is the "real world" anchor moves example.
    "ServerWin_2005.vcproj.Reduced",
    "getCLIArgs.java",
    "chunks",
    "ReleaseNotes",
    "Bug_ReporterApp",
]

# 0 - Perfect.  No issues at all.
# 1 - Conflicts that are handled correctly.
# 2 - Conflicts that are handled correctly.
# 3 - Conflicts that are handled correctly.
# 4 - Technically correct, but SureMerge says they would have flagged it.
# 5 - Too long to tell. (This is the "real world" moves test.)
# 6 - Perfect (Shorter version of real world).
# 7 - Duplicated some code but it's arguably correct.
# 8 - Perfect but it had a weird split.
# 9 - Perfect, but it easily have gone otherwise.


EXCEPTIONS = {
    "chunks": ("chunks_1_0.c", "chunks_1_1.c", "chunks_1_2.c"),
    "ReleaseNotes": (
        "ReleaseNotes 9_4.html",
        "ReleaseNotes 9_4_1.html",
        "ReleaseNotes 9_4_2.html",
    ),
    "Bug_ReporterApp": (
        "Bug_ReporterApp_Parent.h",
        "Bug_ReporterApp_BranchA.h",
        "Bug_ReporterApp_BranchB.h",
    ),
}


def get_filenames(test_case: str):
    if test_case in EXCEPTIONS:
        parent, branch0, branch1 = EXCEPTIONS[test_case]
    else:
        parent = f"{test_case}.parent"
        branch0 = f"{test_case}.1st"
        branch1 = f"{test_case}.2nd"

    parent = os.path.join(TEST_CASE_DIRECTORY, parent)
    branch0 = os.path.join(TEST_CASE_DIRECTORY, branch0)
    branch1 = os.path.join(TEST_CASE_DIRECTORY, branch1)
    return parent, branch0, branch1


def get_usage():
    return (
        f"{PROGRAM_NAME} test_case\nValid test cases are:\n\t"
        + "\n\t".join(TEST_CASES)
        + "\n"
    )



from collections.abc import Iterable, Sequence
from collections import namedtuple

TokenSequence = Sequence[int]
Change = namedtuple("Change", ["a", "b"])

ChangeIterable = Iterable[str | TokenSequence | Change]
StringIterable = Iterable[str]
TokenSequenceIterable = Iterable[TokenSequence]



from typing import Iterable

import gaspra.suffix_automaton as sa


def find_substring(search_in: str, search_for: str) -> Iterable[int]:
    """Find all occurences of search_for in search_in

    Arguments:
        search_in: str
            The string to be searched
        search_for: str
            The string to search for

    Returns
        Iterable[int]
            Positions of every occurrence of `search_for` in `seach_in`.
    """

    root = sa.build(search_in)
    return sa.find_substring_all(root, search_for)


def find_lcs(a: str, b: str):
    """Returns the locations of the first occurences of the LCS of s1 and s2 in
    each string.

    Arguments:
       a: str
          - The first string
       a: str
          - The second string

    Returns:
        int:
           The position of the LCS in s1
        int:
           The position of the LCS in s2
        int:
           The length of the LCS.

    Note that the empty string is always a common substring, occurring at the
    beginning of both lists.
    """
    root = sa.build(a)
    pa, pb, length = sa.find_lcs(root, b)
    return pa, pb, length



import argparse
import os

from gaspra.markup import console_writer, file_writer
from gaspra.markup import (
    GIT_MARKUP,
    SCREEN_MARKUP,
    STRIKEOUT_SCREEN_MARKUP,
    TOKEN_GIT_MARKUP,
    TOKEN_SCREEN_MARKUP,
    line_oriented_markup_changes,
    markup_changes,
    token_oriented_markup_changes,
)

from gaspra.merge import merge
from gaspra.changesets import diff
from gaspra.tokenizers import line_tokenize
from gaspra.types import ChangeIterable
import gaspra.torture_test as torture_test


def add_common_arguments(parser):
    parser.add_argument(
        "-o",
        "--output",
        help="Output file",
        default=None,
    )
    parser.add_argument(
        "-l",
        "--line-oriented",
        action="store_true",
        help="Use a line-oriented diff for output, even if the processing was character-oriented",
    )
    parser.add_argument(
        "-c",
        "--character-oriented",
        action="store_true",
        help="Use a character-oriented diff as opposed to line-oriented.",
    )
    parser.add_argument(
        "-f",
        "--file-style",
        action="store_true",
        help="Use file-oriented vs screen-oriented markup (git-style, no color)",
    )
    parser.add_argument(
        "-s",
        "--strikeout",
        action="store_true",
        help="Use a strikeout font for deletions (only on diffs).",
    )


def get_merge_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("parent")
    parser.add_argument("from_branch_head")
    parser.add_argument("into_branch_head")

    add_common_arguments(parser)

    parser.add_argument(
        "-d", "--diff", action="store_true", help="Show diffs along with merge result"
    )

    args = parser.parse_args()
    return args


def get_diff_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("original")
    parser.add_argument("modified")

    add_common_arguments(parser)

    args = parser.parse_args()
    return args


def get_torture_test_arguments():
    parser = argparse.ArgumentParser(usage=torture_test.get_usage())
    parser.add_argument("test_case")
    add_common_arguments(parser)

    parser.add_argument(
        "-d", "--diff", action="store_true", help="Show diffs along with merge result"
    )

    args = parser.parse_args()
    return args


def get_markup_function(arguments, token_map=(), allow_strikeout=True):
    if not arguments.character_oriented:
        wrapped_markup_function = token_oriented_markup_changes

    elif arguments.line_oriented:
        wrapped_markup_function = line_oriented_markup_changes
    else:
        wrapped_markup_function = markup_changes

    markup = get_markup_style(arguments, allow_strikeout)

    def markup_function(
        writer,
        changeset: ChangeIterable,
        branch0: str,
        branch1: str,
        header: str | None = "",
    ):
        wrapped_markup_function(
            writer,
            changeset,
            os.path.basename(branch0),
            os.path.basename(branch1),
            markup=markup,
            header=os.path.basename(header) if header else None,
            token_map=token_map,
        )

    return markup_function


def get_markup_style(arguments, allow_strikeout=True):
    if not arguments.character_oriented:
        if arguments.file_style:
            return TOKEN_GIT_MARKUP
        return TOKEN_SCREEN_MARKUP
    if arguments.file_style:
        return GIT_MARKUP
    elif arguments.strikeout and allow_strikeout:
        return STRIKEOUT_SCREEN_MARKUP
    return SCREEN_MARKUP


def get_writer(arguments):
    if arguments.output is None:
        return console_writer()

    else:
        return file_writer(arguments.output)


def merge_cli():
    arguments = get_merge_arguments()
    parent_name = arguments.parent
    current_name = arguments.into_branch_head
    other_name = arguments.from_branch_head

    _merge(parent_name, current_name, other_name, arguments)


def torture_cli():
    arguments = get_torture_test_arguments()

    test_case = arguments.test_case

    parent_name, current_name, other_name = torture_test.get_filenames(test_case)

    _merge(parent_name, current_name, other_name, arguments)


def _merge(parent_name, current_name, other_name, arguments):
    parent, current, other = get_text(parent_name, current_name, other_name)

    token_map = None
    if not arguments.character_oriented:
        token_map, parent, current, other = line_tokenize(parent, current, other)

    with get_writer(arguments) as writer:
        if arguments.diff:
            diff_markup = get_markup_function(
                arguments, token_map, allow_strikeout=True
            )
            current_changes = diff(parent, current)
            other_changes = diff(parent, other)

            def markup_one(changes, branch_name):
                diff_markup(
                    writer,
                    changes,
                    branch_name,
                    parent_name,
                    header=branch_name,
                )

            markup_one(current_changes, current_name)
            markup_one(other_changes, other_name)

        merged = merge(parent, current, other)
        merge_markup = get_markup_function(arguments, token_map, allow_strikeout=False)

        merge_markup(
            writer,
            merged,
            current_name,
            other_name,
            header="Merged" if arguments.diff else None,
        )


def diff_cli():
    arguments = get_diff_arguments()

    original_name = arguments.original
    modified_name = arguments.modified

    original, modified = get_text(original_name, modified_name)

    token_map = None
    if not arguments.character_oriented:
        token_map, original, modified = line_tokenize(original, modified)

    changes = diff(original, modified)

    display_function = get_markup_function(arguments, token_map)

    with get_writer(arguments) as writer:
        display_function(
            writer,
            changes,
            modified_name,
            original_name,
        )


def get_text(*filenames: str):
    data = []
    for filename in filenames:
        try:
            with open(filename, "rt", encoding="utf-8") as f:
                data.append(f.read())

        except UnicodeDecodeError:
            with open(filename, "rt", encoding="iso-8859-1") as f:
                data.append(f.read())

    return tuple(data)


if __name__ == "__main__":
    torture_cli()
